/* **************************************************************** *
 *                                                                  *
 *  APRSG-NG -- 2nd generation receive-only APRS-i-gate with        *
 *              minimal requirement of esoteric facilities or       *
 *              libraries of any kind beyond UNIX system libc.      *
 *                                                                  *
 * (c) Matti Aarnio - OH2MQK,  2007                                 *
 *                                                                  *
 * **************************************************************** */


#include "aprsg.h"


/* The ttyreader does read TTY ports into a big buffer, and then from there
   to packet frames depending on what is attached...  */

struct serialport ttys[MAXTTYS];
int serialcount; /* How many are defined ? */


/*
 *  ttyread_kissprocess()  --  the S->rdline[]  array has a KISS frame after
 *  KISS escape decode.  The frame begins with KISS command byte, then
 *  AX25 headers and payload, and possibly a CRC-checksum.
 *  Frame length is in S->rdlinelen variable.
 */

/* KA9Q describes the KISS frame format as follows:

   http://www.ka9q.net/papers/kiss.html

   - - - - - - - - -
   4. Control of the KISS TNC

   To distinguish between command and data frames on the host/TNC link,
   the first byte of each asynchronous frame between host and TNC is
   a "type" indicator.

   This type indicator byte is broken into two 4-bit nibbles so that
   the low-order nibble indicates the command number (given in the table
   below) and the high-order nibble indicates the port number for that
   particular command. In systems with only one HDLC port, it is by definition
   Port 0. In multi-port TNCs, the upper 4 bits of the type indicator
   byte can specify one of up to sixteen ports. The following commands
   are defined in frames to the TNC (the "Command" field is in hexadecimal):

   . . . . . .

   CMD code 0 is for the data frame, and is only one present coming from
   TNC to host.

   - - - - - - - - -

   SYMEK et al. have defined a way to run CRC inside KISS frames to
   verify that the KISS-frame itself is correct:

   http://www.symek.com/g/smack.html
   http://www.ir3ip.net/iw3fqg/doc/smak.htm

   SMACK variation recycles the top-most bit of the TNC-id nibble, and
   thus permits up to 8 TNC ports on line.  Top-most bit is always one
   on SMACK frames.

*/


int  crc16_calc(char *buf, int n)
{

  static int  crc_table[] = {
    0x0000, 0xc0c1, 0xc181, 0x0140, 0xc301, 0x03c0, 0x0280, 0xc241,
    0xc601, 0x06c0, 0x0780, 0xc741, 0x0500, 0xc5c1, 0xc481, 0x0440,
    0xcc01, 0xcc0,  0x0d80, 0xcd41, 0x0f00, 0xcfc1, 0xce81, 0x0e40,
    0x0a00, 0xcac1, 0xcb81, 0x0b40, 0xc901, 0x09c0, 0x0880, 0xc841,
    0xd801, 0x18c0, 0x1980, 0xd941, 0x1b00, 0xdbc1, 0xda81, 0x1a40,
    0x1e00, 0xdec1, 0xdf81, 0x1f40, 0xdd01, 0x1dc0, 0x1c80, 0xdc41,
    0x1400, 0xd4c1, 0xd581, 0x1540, 0xd701, 0x17c0, 0x1680, 0xd641,
    0xd201, 0x12c0, 0x1380, 0xd341, 0x1100, 0xd1c1, 0xd081, 0x1040,
    0xf001, 0x30c0, 0x3180, 0xf141, 0x3300, 0xf3c1, 0xf281, 0x3240,
    0x3600, 0xf6c1, 0xf781, 0x3740, 0xf501, 0x35c0, 0x3480, 0xf441,
    0x3c00, 0xfcc1, 0xfd81, 0x3d40, 0xff01, 0x3fc0, 0x3e80, 0xfe41,
    0xfa01, 0x3ac0, 0x3b80, 0xfb41, 0x3900, 0xf9c1, 0xf881, 0x3840,
    0x2800, 0xe8c1, 0xe981, 0x2940, 0xeb01, 0x2bc0, 0x2a80, 0xea41,
    0xee01, 0x2ec0, 0x2f80, 0xef41, 0x2d00, 0xedc1, 0xec81, 0x2c40,
    0xe401, 0x24c0, 0x2580, 0xe541, 0x2700, 0xe7c1, 0xe681, 0x2640,
    0x2200, 0xe2c1, 0xe381, 0x2340, 0xe101, 0x21c0, 0x2080, 0xe041,
    0xa001, 0x60c0, 0x6180, 0xa141, 0x6300, 0xa3c1, 0xa281, 0x6240,
    0x6600, 0xa6c1, 0xa781, 0x6740, 0xa501, 0x65c0, 0x6480, 0xa441,
    0x6c00, 0xacc1, 0xad81, 0x6d40, 0xaf01, 0x6fc0, 0x6e80, 0xae41,
    0xaa01, 0x6ac0, 0x6b80, 0xab41, 0x6900, 0xa9c1, 0xa881, 0x6840,
    0x7800, 0xb8c1, 0xb981, 0x7940, 0xbb01, 0x7bc0, 0x7a80, 0xba41,
    0xbe01, 0x7ec0, 0x7f80, 0xbf41, 0x7d00, 0xbdc1, 0xbc81, 0x7c40,
    0xb401, 0x74c0, 0x7580, 0xb541, 0x7700, 0xb7c1, 0xb681, 0x7640,
    0x7200, 0xb2c1, 0xb381, 0x7340, 0xb101, 0x71c0, 0x7080, 0xb041,
    0x5000, 0x90c1, 0x9181, 0x5140, 0x9301, 0x53c0, 0x5280, 0x9241,
    0x9601, 0x56c0, 0x5780, 0x9741, 0x5500, 0x95c1, 0x9481, 0x5440,
    0x9c01, 0x5cc0, 0x5d80, 0x9d41, 0x5f00, 0x9fc1, 0x9e81, 0x5e40,
    0x5a00, 0x9ac1, 0x9b81, 0x5b40, 0x9901, 0x59c0, 0x5880, 0x9841,
    0x8801, 0x48c0, 0x4980, 0x8941, 0x4b00, 0x8bc1, 0x8a81, 0x4a40,
    0x4e00, 0x8ec1, 0x8f81, 0x4f40, 0x8d01, 0x4dc0, 0x4c80, 0x8c41,
    0x4400, 0x84c1, 0x8581, 0x4540, 0x8701, 0x47c0, 0x4680, 0x8641,
    0x8201, 0x42c0, 0x4380, 0x8341, 0x4100, 0x81c1, 0x8081, 0x4040
  };

  int  crc;

  crc = 0;
  while (--n >= 0)
    crc = ((crc >> 8) & 0xff) ^ crc_table[(crc ^ *buf++) & 0xff];
  return crc;
}



static int ttyread_kissprocess(struct serialport *S)
{
	int i;
	int cmdbyte;

	cmdbyte = S->rdline[0];

	// --
	// C0 00
	// 82 A0 B4 9A 88 A4 60
	// 9E 90 64 90 A0 9C 72
	// 9E 90 64 A4 88 A6 E0
	// A4 8C 9E 9C 98 B2 61
	// 03 F0
	// 21 36 30 32 39 2E 35 30 4E 2F 30 32 35 30 35 2E 34 33 45 3E 20 47 43 53 2D 38 30 31 20
	// C0
	// --

	// printf("ttyread_kissprocess()  cmdbyte=%02X len=%d ",cmdbyte,S->rdlinelen);

	/* Ok, cmdbyte tells us something, and we should ignore the
	   frame if we don't know it... */

	if ((cmdbyte & 0x0F) != 0) {
	  /* There should NEVER be any other value in the CMD bits
	     than 0  coming from TNC to host! */
	  // printf(" ..bad CMD byte\n");
	  return -1;
	}

	/* Are we excepting AEA "CRC" */
	if (S->linetype == LINETYPE_KISSCRCAEA) {
	}
	/* Are we expecting SMACK ? */
	if ((S->linetype == LINETYPE_KISSCRC) &&
	    ((cmdbyte & 0x8F) == 0x80)) {

	  /* It is SMACK frame -- KISS with CRC16 at the tail.
	     Now we ignore the TNC-id number field.
	     Verify the CRC.. */

	  if (crc16_calc(S->rdline+1, S->rdlinelen-1) != 0) {
	    // printf(" ..invalid CRC value\n");
	    return -1;   // The CRC was invalid..
	  }

	  S->rdlinelen -= 2;  /* Chop off the two CRC bytes */
	}

	if (S->rdlinelen < 17) {
	  /* 7+7+2 bytes of minimal AX.25 frame + 1 for KISS CMD byte */

	  /* Too short frame.. */
	  // printf(" ..too short a frame for anything\n");
	  return -1;
	}

	/* Valid AX.25 HDLC frame byte sequence is now at
	       S->rdline[1..S->rdlinelen-1]
	*/

	/* ADD?  send(somesocket, S->rdline+1, S->rdlinelen-1)  ??
	   Send the AX.25 frame to some socket for other processing uses.. */

	// printf("\n");

	ax25_to_tnc2(cmdbyte, S->rdline+1, S->rdlinelen-1);

	return 0;
}


/*
 *  ttyreader_getc()  -- pick one char ( >= 0 ) out of input buffer, or -1 if out of buffer
 */
static int ttyreader_getc(struct serialport *S)
{
	if (S->rdcursor >= S->rdlen) {  /* Out of data ? */
	  if (S->rdcursor)
	    S->rdcursor = S->rdlen = 0;
	  // printf("-\n");
	  return -1;
	}

	// printf(" %02X", 0xFF & S->rdbuf[S->rdcursor++]);

	return (0xFF & S->rdbuf[S->rdcursor++]);
}


/*
 * ttyreader_pullkiss()  --  pull KISS (or KISS+CRC) frame, and call KISS processor
 */

#define KISS_FEND  (0xC0)
#define KISS_FESC  (0xDB)
#define KISS_TFEND (0xDC)
#define KISS_TFESC (0xDD)

static int ttyreader_pullkiss(struct serialport *S)
{
	// printf("ttyreader_pullkiss()  rdlen=%d rdcursor=%d, state=%d\n", S->rdlen, S->rdcursor, S->kissstate); fflush(stdout);

	/* At incoming call there is at least one byte in between
	   S->rdcursor and S->rdlen  */

	/* Phases:
	   kissstate == 0: hunt for KISS_FEND, discard everything before it.
	   kissstate != 0: reading has globbed up preceding KISS_FENDs ("HDLC flags")
	                   and the cursor is in front of a frame
	*/

	/* There are TNCs that use "shared flags" - only one FEND in between data frames. */

	if (S->kissstate == KISSSTATE_SYNCHUNT) {
	  /* Hunt for KISS_FEND, discard everything until then! */
	  int i;
	  int c;
	  for (;;) {
	    c = ttyreader_getc(S);
	    if (c < 0) return c;    /* Out of buffer, stay in state,
				     return latter when there is some refill */
	    if (c == KISS_FEND)   /* Found the sync-byte !  change state! */
	      break;
	  }
	  S->kissstate = KISSSTATE_COLLECTING;
	}


	if (S->kissstate != KISSSTATE_SYNCHUNT) {
	  /* Normal processing mode */

	  int c;

	  for (;;) {
	    c = ttyreader_getc(S);
	    if (c < 0) return c; /* Out of input stream, exit now, come back latter.. */

	    // printf(" %02X", c);
	    // if (c == KISS_FEND) { printf("\n");fflush(stdout); }

	    if (c == KISS_FEND) {
	      /* Found end-of-frame character -- or possibly beginning..
		 This never exists in datastream except as itself. */

	      if (S->rdlinelen > 0) {
		/* Non-zero sized frame  Process it away ! */
		ttyread_kissprocess(S);
		S->kissstate = KISSSTATE_COLLECTING;
		S->rdlinelen = 0;
	      }

	      /* rdlinelen == 0 because we are receiving consequtive
		 FENDs, or just processed our previous frame.  Treat
		 them the same: discard this byte. */

	      continue;
	    }

	    if (S->kissstate == KISSSTATE_KISSFESC) {

	      /* We have some char, state switches to normal collecting */
	      S->kissstate = KISSSTATE_COLLECTING;

	      if      (c == KISS_TFEND)   c = KISS_FEND;
	      else if (c == KISS_TFESC)   c = KISS_FESC;
	      else
		continue;  /* Accepted chars after KISS_FESC
			      are only TFEND and TFESC.
			      Others must be discarded. */

	    } else { /* Normal collection mode */

	      if (c == KISS_FESC) {
		S->kissstate = KISSSTATE_KISSFESC;
		continue;  /* Back to top of the loop and continue.. */
	      }

	    }


	    if (S->rdlinelen >= (sizeof(S->rdline)-3)) { /* Too long !  Way too long ! */
	      S->kissstate = KISSSTATE_SYNCHUNT;  /* Sigh.. discard it. */
	      S->rdlinelen = 0;
	      continue;
	    }

	    /* Put it on record store: */
	    S->rdline[S->rdlinelen++] = c;
	  } /* .. for(..) loop of data collecting */

	} /* .. normal consumption mode ... */

	return 0;
}


/*
 *  ttyreader_pulltnc2()  --  process a line of text by calling TNC2 UI Monitor analyzer
 */

static int ttyreader_pulltnc2(struct serialport *S)
{
	int i;

	
	return 0;
}

/*
 * ttyreader_pullaea()  --  process a line of text by calling AEA MONITOR 1 analyzer
 */

static int ttyreader_pullaea(struct serialport *S)
{
	int i;

	if (S->rdline[S->rdlinelen-1] == ':') {
	  /* Could this be the AX25 header ? */
	  char *s = strchr(S->rdline, '>');
	  if (s) {
	    /* Ah yes, it well could be.. */
	    strcpy(S->rdline2, S->rdline);
	    return;
	  }
	}

	/* FIXME: re-arrange the  S->rdline2  contained AX25 address tokens and flags..
	   perl code:
	      @addrs = split('>', $rdline2);
	      $out = shift @addrs; # pop first token in sequence
	      $out .= '>';
	      $out .= pop @addrs;  # pop last token in sequence
	      foreach $a (@addrs) { # rest of the tokens in sequence, if any
	          $out .= ',' . $a;
	      }
	      # now $out has address data in TNC2 sequence.
	*/

	// printf("%s%s\n", S->rdline2, S->rdline); fflush(stdout);

	return 0;
}





/*
 *  ttyreader_pulltext()  -- process a line of text from the serial port..
 */

static int ttyreader_pulltext(struct serialport *S)
{
	int c;

	for (;;) {

	  c = ttyreader_getc(S);
	  if (c < 0) return c;   /* Out of input.. */

	  /* S->kissstate != 0: read data into S->rdline, == 0: discard data until CR|LF.
	     Zero-size read line is discarded as well (only CR|LF on input frame)  */

	  if (S->kissstate == KISSSTATE_SYNCHUNT) {
	    /* Looking for CR or LF.. */
	    if (c == '\n' || c == '\r')
	      S->kissstate = KISSSTATE_COLLECTING;

	    S->rdlinelen = 0;
	    continue;
	  }

	  /* Now: (S->kissstate != KISSSTATE_SYNCHUNT)  */

	  if (c == '\n' || c == '\r') {
	    /* End of line seen! */
	    if (S->rdlinelen > 0) {

	      /* Non-zero-size string, put terminating 0 byte on it. */
	      S->rdline[S->rdlinelen] = 0;

	      /* .. and process it depending ..  */

	      if (S->linetype == LINETYPE_TNC2) {
		ttyreader_pulltnc2(S);
	      } else {  /* .. it is LINETYPE_AEA ? */
		ttyreader_pullaea(S);
	      }
	    }
	    S->rdlinelen = 0;
	    continue;
	  }

	  /* Now place the char in the linebuffer, if there is space.. */
	  if (S->rdlinelen >= (sizeof(S->rdline)-3)) { /* Too long !  Way too long ! */
	    S->kissstate = KISSSTATE_SYNCHUNT;  /* Sigh.. discard it. */
	    S->rdlinelen = 0;
	    continue;
	  }

	  /* Put it on line store: */
	  S->rdline[S->rdlinelen++] = c;

	} /* .. input loop */

	return 0; /* not reached */
}



/*
 *  ttyreader_lineread()  --  read what there is into our buffer, and process the buffer..
 */

static void ttyreader_lineread(struct serialport *S)
{
	int i;

	int rdspace = sizeof(S->rdbuf) - S->rdlen;

	if (S->rdcursor > 0) {
	  /* Read-out cursor is not at block beginning,
	     is there unread data too ?  */
	  if (S->rdlen > S->rdcursor) {
	    /* Uh..  lets move buffer down a bit,
	       to make room for more to the end.. */
	    memcpy(S->rdbuf, S->rdbuf + S->rdcursor, 
		   S->rdlen - S->rdcursor);
	    S->rdlen = S->rdlen - S->rdcursor;
	  } else
	    S->rdlen = 0; /* all processed, mark its size zero */
	  /* Cursor to zero, rdspace recalculated */
	  S->rdcursor = 0;

	  /* recalculate */
	  rdspace = sizeof(S->rdbuf) - S->rdlen;
	}

	if (rdspace > 0) { /* We have room to read into.. */
	  i = read(S->fd, S->rdbuf + S->rdlen, rdspace);
	  if (i == 0) { /* EOF ?  USB unplugged ? */
	    close(S->fd);
	    S->fd = -1;
	    S->wait_until = now + TTY_OPEN_RETRY_DELAY_SECS;
	    return;
	  }
	  if (i < 0) /* EAGAIN or whatever.. */
	    return;

	  /* Some data has been accumulated ! */
	  S->rdlen += i;
	}

	/* Done reading, maybe.  Now processing.
	   The pullXX does read up all input, and does
	   however many frames there are in, and pauses
	   when there is no enough input data for a full
	   frame/line/whatever.
	*/

	if (S->linetype == LINETYPE_KISS ||
	    S->linetype == LINETYPE_KISSCRC) {

	  ttyreader_pullkiss(S);

	} else if (S->linetype == LINETYPE_TNC2 ||
		   S->linetype == LINETYPE_AEA) {

	  ttyreader_pulltext(S);

	} else {
	  close(S->fd); /* Urgh ?? Bad linetype value ?? */
	  S->fd = -1;
	  S->wait_until = now + TTY_OPEN_RETRY_DELAY_SECS;
	}

	/* Consumed something, and our read cursor is not in the beginning ? */
	if (S->rdcursor > 0 && S->rdcursor < S->rdlen) {
	  /* Compact the input buffer! */
	  memmove(S->rdbuf, S->rdbuf + S->rdcursor, S->rdlen - S->rdcursor);
	}
	S->rdlen -= S->rdcursor;
	S->rdcursor = 0;
}


/*
 * ttyreader_linesetup()  --  open and configure the serial port
 */

static void ttyreader_linesetup(struct serialport *S)
{
	int i;
	const char *s;

	S->wait_until = 0; /* Zero it just to be safe */

	S->fd = open(S->ttyname, O_RDWR|O_NOCTTY|O_NONBLOCK, 0);
	if (S->fd < 0) { /* Urgh.. an error.. */
	  S->wait_until = now + TTY_OPEN_RETRY_DELAY_SECS;
	  return;
	}

	i = tcsetattr(S->fd, TCSANOW, &S->tio);
	if (i < 0) {
	  close(S->fd);
	  S->fd = -1;
	  S->wait_until = now + TTY_OPEN_RETRY_DELAY_SECS;
	  return;
	}
	/* FIXME: ??  Set baud-rates ?
	   Used system (Linux) has them in   'struct termios'..
	 */

	switch (S->linetype) {
	case LINETYPE_KISS:
	case LINETYPE_KISSCRC: // SMACK ...
	  s = "\xC0\xC0\xFF\xC0\r\rMO 0\rKISS $01\r";
	  break;
	case LINETYPE_KISSCRCAEA:
	  s = "\xC0\xC0\xFF\xC0\r\rMO 0\rKISS $0B\r";
	  break;
	case LINETYPE_TNC2:
	case LINETYPE_AEA:
	  s = "\xC0\xC0\xFF\xC0\xC0\xFF\xC0\xC0\r\r\rMO 1\r";
	  break;
	default:
	  break;
	}
	write(S->fd, s, strlen(s));


	S->rdlen = S->rdcursor = S->rdlinelen = 0;
	S->kissstate = 0; /* Zero it, whatever protocol we actually use
			     will consider it as 'hunt for sync' state. */
}


/*
 *  ttyreader_prepoll()  --  prepare system for next round of polling
 */

int ttyreader_prepoll(int nfds, struct pollfd **fdsp, time_t *tout)
{
	int idx = 0; /* returns number of *fds filled.. */
	int i;
	struct serialport *S = ttys;
	struct pollfd *fds = *fdsp;

	for (i = 0; i < MAXTTYS; ++i, ++S) {
	  if (*S->ttyname == 0) continue; /* No name, no look... */
	  if (S->fd < 0) {
	    /* Not an open TTY, but perhaps waiting ? */
	    if ((S->wait_until != 0) && (S->wait_until > now)) {
	      /* .. waiting for future! */
	      if (*tout > S->wait_until)
		*tout = S->wait_until;
	      /* .. but only until our timeout,
		 if it is sooner than global one. */
	      continue; /* Waiting on this one.. */
	    }

	    /* Waiting or not, FD is not open, and deadline is past.
	       Lets try to open! */

	    ttyreader_linesetup(S);

	  } /* .. No open FD */
	  /* Still no open FD ? */
	  if (S->fd < 0)
	    continue;

	  /* FD is open, lets mark it for poll read.. */
	  fds->fd = S->fd;
	  fds->events = POLLIN|POLLPRI;
	  fds->revents = 0;
	  ++fds;
	  ++idx;
	  if (idx >= nfds) /* consumed all than there was ? */
	    break;
	}
	*fdsp = fds;
	return idx;
}


/*
 *  ttyreader_postpoll()  -- Done polling, what happened ?
 */

int ttyreader_postpoll(int nfds, struct pollfd *fds)
{
	int idx, i;

	struct serialport *S;
	struct pollfd     *P;
	for (idx = 0, P = fds; idx < nfds; ++idx, ++P) {
	  if (! (P->revents & (POLLIN | POLLPRI | POLLERR | POLLHUP)))
	    continue; /* No read event we are interested in... */

	  for (i = 0, S = ttys; i < MAXTTYS; ++i, ++S) {
	    if (S->fd != P->fd) continue; /* Not this one ? */
	    /* It is this one! */
	    ttyreader_lineread(S);
	  }
	}
	return 0;
}
